@@ -23,10 +23,15 @@ extern "C" {
  * @author Thomas Hallgren
  *******************************************************************/
 
+/*
+ * Create a Type for a specific Record type.
+ */
+extern Type SingleRowWriter_createType(Oid typid, TupleDesc tupleDesc);
+
 /*
  * Create an instance of org.postgresql.pljava.jdbc.RowProviderSet
  */
-extern jobject SingleRowWriter_create(JNIEnv* env, TupleDesc tupleDesc);
+extern jobject SingleRowWriter_create(JNIEnv* env, jobject tupleDesc);
 
 /*
  * Returns the Tuple for the SingleRowWriter and clears it.
