@@ -16,6 +16,11 @@
 #include <fmgr.h>
 #include <access/heapam.h>
 #include <utils/syscache.h>
+#include <catalog/catalog.h>
+#if (PGSQL_MAJOR_VER < 8)
+#	include <fcntl.h>
+#	include <catalog/pg_control.h>
+#endif
 #include <catalog/pg_proc.h>
 #include <catalog/pg_type.h>
 #include <storage/ipc.h>
@@ -37,7 +42,6 @@
 #include "pljava/SPI.h"
 #include "pljava/type/NativeStruct.h"
 #include "pljava/type/String.h"
-
 /* Example format: "/usr/local/pgsql/lib" */
 #ifndef PKGLIBDIR
 #error "PKGLIBDIR needs to be defined to compile this file."
@@ -49,6 +53,9 @@ bool isCallingJava;
 
 static JNIEnv* s_mainEnv = 0;
 static JavaVM* s_javaVM = 0;
+static jclass  s_Backend_class;
+static jmethodID s_setTrusted;
+static bool    s_currentTrust = false;
 
 #if (PGSQL_MAJOR_VER >= 8)
 # define PGSQL_CUSTOM_VARIABLES 1
@@ -108,6 +115,11 @@ static void initPLJavaClasses(JNIEnv* env)
 
 
 	JNINativeMethod invocationMethods[] = {
+		{
+		"_getCurrent",
+		"()Lorg/postgresql/pljava/jdbc/Invocation;",
+		Java_org_postgresql_pljava_jdbc_Invocation__1getCurrent
+		},
 		{
 		"_getNestingLevel",
 		"()I",
@@ -125,7 +137,9 @@ static void initPLJavaClasses(JNIEnv* env)
 		},
 		{ 0, 0, 0 }};
 
-	PgObject_registerNatives(env, "org/postgresql/pljava/internal/Backend", backendMethods);
+	s_Backend_class = PgObject_getJavaClass(env, "org/postgresql/pljava/internal/Backend");
+	PgObject_registerNatives2(env, s_Backend_class, backendMethods);
+	s_setTrusted = PgObject_getStaticJavaMethod(env, s_Backend_class, "setTrusted", "(Z)V");
 
 	cls = PgObject_getJavaClass(env, "org/postgresql/pljava/jdbc/Invocation");
 	PgObject_registerNatives2(env, cls, invocationMethods);
@@ -142,31 +156,82 @@ static void initPLJavaClasses(JNIEnv* env)
 static bool s_topLocalFrameInstalled = false;
 static unsigned int s_callLevel = 0;
 
-CallContext* currentCallContext;
-
-static void initCallContext(CallContext* ctx)
-{
-	ctx->invocation    = 0;
-	ctx->function      = 0;
-	ctx->upperContext  = CurrentMemoryContext;
-	ctx->errorOccured  = false;
-	ctx->previous      = currentCallContext;
-	currentCallContext = ctx;
-}
-
 static void popJavaFrameCB(MemoryContext ctx, bool isDelete)
 {
 	if(s_callLevel == 0 && s_topLocalFrameInstalled)
 	{
 		/* Pop this frame. This might call finalizers.
 		 */
-		JNIEnv* env = Backend_getMainEnv();
+		JNIEnv* env = Backend_getJNIEnv();
 		if(env != 0)
 			Backend_popJavaFrame(env);
 		s_topLocalFrameInstalled = false;
 	}
 }
 
+/**
+ *  Initialize security
+ */
+static void setJavaSecurity(JNIEnv* env, bool trusted)
+{
+/* GCJ has major issues here. Real work on SecurityManager and
+ * related classes has just started in version 4.0.0.
+ */
+#ifndef GCJ
+	bool saveICJ = isCallingJava;
+	isCallingJava = true;
+	(*env)->CallStaticVoidMethod(env, s_Backend_class, s_setTrusted, (jboolean)trusted);
+	isCallingJava = saveICJ;
+
+	if((*env)->ExceptionCheck(env))
+	{
+		(*env)->ExceptionDescribe(env);
+		(*env)->ExceptionClear(env);
+		ereport(ERROR, (
+			errcode(ERRCODE_INTERNAL_ERROR),
+			errmsg("Unable to initialize java security")));
+	}
+#endif
+}
+
+bool integerDateTimes = false;
+CallContext* currentCallContext;
+
+void Backend_pushCallContext(CallContext* ctx, bool trusted)
+{
+	ctx->jniEnv          = s_mainEnv;	/* The one and only at the moment */
+	ctx->invocation      = 0;
+	ctx->function        = 0;
+	ctx->trusted         = trusted;
+	ctx->hasConnected    = false;
+	ctx->upperContext    = CurrentMemoryContext;
+	ctx->errorOccured    = false;
+	ctx->inExprContextCB = false;
+	ctx->previous        = currentCallContext;
+	currentCallContext   = ctx;
+
+	if(trusted != s_currentTrust)
+	{
+		setJavaSecurity(ctx->jniEnv, trusted);
+		s_currentTrust = trusted;
+	}
+}
+
+void Backend_popCallContext(void)
+{
+	CallContext* ctx = currentCallContext->previous;
+	if(ctx != 0)
+	{
+		if(ctx->trusted != s_currentTrust)
+		{
+			setJavaSecurity(ctx->jniEnv, ctx->trusted);
+			s_currentTrust = ctx->trusted;
+		}
+		MemoryContextSwitchTo(ctx->upperContext);
+	}
+	currentCallContext = ctx;
+}
+
 void Backend_pushJavaFrame(JNIEnv* env)
 {
 	if((*env)->PushLocalFrame(env, LOCAL_REFERENCE_COUNT) < 0)
@@ -193,7 +258,7 @@ void Backend_popJavaFrame(JNIEnv* env)
 
 bool pljavaEntryFence(JNIEnv* env)
 {
-	if(currentCallContext->errorOccured)
+	if(currentCallContext != 0 && currentCallContext->errorOccured)
 	{
 		/* An elog with level higher than ERROR was issued. The transaction
 		 * state is unknown. There's no way the JVM is allowed to enter the
@@ -421,8 +486,6 @@ static void _destroyJavaVM(int status, Datum dummy)
 		enable_sig_alarm(5000, false);
 #endif
 
-		initCallContext(&ctx);
-
 		elog(DEBUG1, "Destroying JavaVM...");
 
 		isCallingJava = true;
@@ -570,22 +633,67 @@ static void addUserJVMOptions(JVMOptList* optList)
 }
 #endif
 
-
 /**
- *  Initialize the logger
+ *  Initialize the session
  */
-static void initJavaLogger(JNIEnv* env)
+static void initJavaSession(JNIEnv* env)
 {
-	jclass logHandlerClass;
+	jclass sessionClass;
 	jmethodID init;
 	bool saveICJ = isCallingJava;
 
 	isCallingJava = true;
-	logHandlerClass = PgObject_getJavaClass(env, "org/postgresql/pljava/internal/ELogHandler");
-	init = PgObject_getStaticJavaMethod(env, logHandlerClass, "init", "()V");
-	(*env)->CallStaticVoidMethod(env, logHandlerClass, init);
+	sessionClass = PgObject_getJavaClass(env, "org/postgresql/pljava/internal/Session");
+	init = PgObject_getStaticJavaMethod(env, sessionClass, "init", "()V");
+	(*env)->CallStaticVoidMethod(env, sessionClass, init);
 	isCallingJava = saveICJ;
-	(*env)->DeleteLocalRef(env, logHandlerClass);
+	(*env)->DeleteLocalRef(env, sessionClass);
+
+	if((*env)->ExceptionCheck(env))
+	{
+		(*env)->ExceptionDescribe(env);
+		(*env)->ExceptionClear(env);
+		ereport(ERROR, (
+			errcode(ERRCODE_INTERNAL_ERROR),
+			errmsg("Unable to initialize java session")));
+	}
+}
+
+static void checkIntTimeType(void)
+{
+#if (PGSQL_MAJOR_VER >= 8)
+	const char* idt = GetConfigOption("integer_datetimes");
+	integerDateTimes = (strcmp(idt, "on") == 0);
+#else	
+	ControlFileData data;
+	int   fd;
+	char  controlFilePath[MAXPGPATH];
+	char* globalTableSpace = GetDatabasePath((Oid)0);
+
+	snprintf(controlFilePath, MAXPGPATH, "%s/pg_control", globalTableSpace);
+	pfree(globalTableSpace);
+
+	if((fd = open(controlFilePath, O_RDONLY)) == -1)
+	{
+		ereport(ERROR, (
+			errcode(ERRCODE_INTERNAL_ERROR),
+			errmsg("could not open file \"%s\" for reading: %s\n",
+				controlFilePath, strerror(errno))));
+	}
+
+	if(read(fd, &data, sizeof(ControlFileData)) != sizeof(ControlFileData))
+	{
+		int errCode = errno;
+		close(fd);
+		ereport(ERROR, (
+			errcode(ERRCODE_INTERNAL_ERROR),
+			errmsg("could not read file \"%s\": %s\n",
+				controlFilePath, strerror(errCode))));
+	}
+	close(fd);
+	integerDateTimes = data.enableIntTimes;
+#endif
+	elog(DEBUG1, integerDateTimes ? "Using integer_datetimes" : "Not using integer_datetimes");
 }
 
 static void initializeJavaVM(void)
@@ -602,6 +710,7 @@ static void initializeJavaVM(void)
 
 	JVMOptList_init(&optList);
 
+	checkIntTimeType();
 	DirectFunctionCall1(HashMap_initialize, 0);
 
 #ifdef PGSQL_CUSTOM_VARIABLES
@@ -634,6 +743,7 @@ static void initializeJavaVM(void)
 		"Size of the prepared statement MRU cache",
 		NULL,
 		&statementCacheSize,
+		0, 512,
 		PGC_USERSET,
 		NULL, NULL);
 
@@ -698,7 +808,7 @@ static void initializeJavaVM(void)
 	{
 		elog(INFO, "Backend pid = %d. Attach the debugger and set pljavaDebug to false to continue", getpid());
 		while(pljavaDebug)
-#if (PGSQL_MAJOR_VER >= 8)
+#if (PGSQL_MAJOR_VER >= 8) && !defined(__OS2__)
 			pg_usleep(1000000L);
 #else
 			sleep(1);
@@ -743,42 +853,82 @@ static void initializeJavaVM(void)
 	 */
 	on_proc_exit(_destroyJavaVM, 0);
 	initPLJavaClasses(s_mainEnv);
-	initJavaLogger(s_mainEnv);
-	if((*s_mainEnv)->ExceptionCheck(s_mainEnv))
+	initJavaSession(s_mainEnv);
+}
+
+void
+Backend_assertConnect(void)
+{
+	if(currentCallContext != 0 && !currentCallContext->hasConnected)
 	{
-		(*s_mainEnv)->ExceptionDescribe(s_mainEnv);
-		(*s_mainEnv)->ExceptionClear(s_mainEnv);
-		ereport(ERROR, (
-			errcode(ERRCODE_INTERNAL_ERROR),
-			errmsg("Unable to initialize java logger")));
+		SPI_connect();
+		currentCallContext->hasConnected = true;
 	}
 }
 
-JNIEnv* Backend_getMainEnv(void)
+void
+Backend_assertDisconnect(void)
 {
-	return s_mainEnv;
+	if(currentCallContext != 0 && currentCallContext->hasConnected)
+	{
+		SPI_finish();
+		currentCallContext->hasConnected = false;
+	}
+}
+
+JNIEnv* Backend_getJNIEnv(void)
+{
+	return (currentCallContext == 0)
+		? s_mainEnv
+		: currentCallContext->jniEnv;
+}
+
+static Datum internalCallHandler(bool trusted, PG_FUNCTION_ARGS);
+
+extern Datum javau_call_handler(PG_FUNCTION_ARGS);
+PG_FUNCTION_INFO_V1(javau_call_handler);
+
+/*
+ * This is the entry point for all untrusted calls.
+ */
+Datum javau_call_handler(PG_FUNCTION_ARGS)
+{
+	return internalCallHandler(false, fcinfo);
 }
 
 extern Datum java_call_handler(PG_FUNCTION_ARGS);
 PG_FUNCTION_INFO_V1(java_call_handler);
 
 /*
- * This is the entry point for all calls. The java_call_handler must be
- * defined in plsql like this:
- *
+ * This is the entry point for all trusted calls.
  */
 Datum java_call_handler(PG_FUNCTION_ARGS)
+{
+	return internalCallHandler(true, fcinfo);
+}
+
+static Datum internalCallHandler(bool trusted, PG_FUNCTION_ARGS)
 {
 	CallContext ctx;
 	Datum retval = 0;
 	bool saveIsCallingJava = isCallingJava;
-	Oid funcOid = fcinfo->flinfo->fn_oid;
-
-	initCallContext(&ctx);
 
 	if(s_javaVM == 0)
+	{
+		/* Initialize the VM, we pass the s_currentTrust here
+		 * to ensure that the pushCallContext doesn't call on
+		 * Java until the JVM is initialized.
+		 */
+		Backend_pushCallContext(&ctx, s_currentTrust);
 		initializeJavaVM();
+		Backend_popCallContext();
 
+		/* Force initial setting
+		 */
+		s_currentTrust = !trusted;
+	}
+
+	Backend_pushCallContext(&ctx, trusted);
 	if(s_callLevel == 0 && !s_topLocalFrameInstalled)
 	{
 		Backend_pushJavaFrame(s_mainEnv);
@@ -786,23 +936,20 @@ Datum java_call_handler(PG_FUNCTION_ARGS)
 		MemoryContext_addEndOfScopeCB(CurrentMemoryContext, popJavaFrameCB);
 	}
 
-	SPI_connect();
-
 	++s_callLevel;
 	PG_TRY();
 	{
+		Function function = Function_getFunction(s_mainEnv, fcinfo);
 		if(CALLED_AS_TRIGGER(fcinfo))
 		{
 			/* Called as a trigger procedure
 			 */
-			Function function = Function_getFunction(s_mainEnv, funcOid, true);
 			retval = Function_invokeTrigger(function, s_mainEnv, fcinfo);
 		}
 		else
 		{
 			/* Called as a function
 			 */
-			Function function = Function_getFunction(s_mainEnv, funcOid, false);
 			retval = Function_invoke(function, s_mainEnv, fcinfo);
 		}
 		Exception_checkException(s_mainEnv);
@@ -814,18 +961,18 @@ Datum java_call_handler(PG_FUNCTION_ARGS)
 		}
 
 		--s_callLevel;
-		isCallingJava      = saveIsCallingJava;
-		currentCallContext = ctx.previous;
-		SPI_finish();
+		isCallingJava = saveIsCallingJava;
+		Backend_assertDisconnect();
+		Backend_popCallContext();
 	}
 	PG_CATCH();
 	{
 		--s_callLevel;
-		isCallingJava      = saveIsCallingJava;
-		currentCallContext = ctx.previous;
+		isCallingJava = saveIsCallingJava;
 		if(ctx.invocation != 0)
 			(*s_mainEnv)->DeleteGlobalRef(s_mainEnv, ctx.invocation);
-		SPI_finish();
+		Backend_assertDisconnect();
+		Backend_popCallContext();
 		PG_RE_THROW();
 	}
 	PG_END_TRY();
@@ -897,11 +1044,41 @@ Java_org_postgresql_pljava_internal_Backend__1getStatementCacheSize(JNIEnv* env,
 JNIEXPORT void JNICALL
 JNICALL Java_org_postgresql_pljava_internal_Backend__1log(JNIEnv* env, jclass cls, jint logLevel, jstring jstr)
 {
-	char* str;
-	str = String_createNTS(env, jstr);
+	int percentCount = 0;
+	const char* cp;
+	char c;
+	char* str = String_createNTS(env, jstr);
 	if(str == 0)
 		return;
 
+	/* elog uses printf formatting but the logger does not so we must escape all
+	 * '%' in the string.
+	 */
+	for(cp = str; (c = *cp) != 0; ++cp)
+	{
+		if(c == '%')
+			++percentCount;
+	}
+
+	if(percentCount > 0)
+	{
+		/* Make room to expand all "%" to "%%"
+		 */
+		char* str2 = palloc((cp - str) + percentCount + 1);
+		char* cp2 = str2;
+
+		/* Expand... */
+		for(cp = str; (c = *cp) != 0; ++cp)
+		{
+			if(c == '%')
+				*cp2++ = c;
+			*cp2++ = c;
+		}
+		*cp2 = 0;
+		pfree(str);
+		str = str2;
+	}
+
 	PG_TRY();
 	{
 		elog(logLevel, str);
@@ -964,16 +1141,7 @@ Java_org_postgresql_pljava_internal_Backend__1addEOXactListener(JNIEnv* env, jcl
 JNIEXPORT void JNICALL
 Java_org_postgresql_pljava_internal_Backend__1removeEOXactListener(JNIEnv* env, jclass cls, jobject listener)
 {
-	PLJAVA_ENTRY_FENCE_VOID
-	PG_TRY();
-	{
-		EOXactListener_unregister(env);
-	}
-	PG_CATCH();
-	{
-		Exception_throw_ERROR(env, "UnregisterEOXactCallback");
-	}
-	PG_END_TRY();
+	EOXactListener_unregister(env);
 }
 
 /*
@@ -987,6 +1155,17 @@ Java_org_postgresql_pljava_jdbc_Invocation__1getNestingLevel(JNIEnv* env, jclass
 	return s_callLevel;
 }
 
+/*
+ * Class:     org_postgresql_pljava_jdbc_Invocation
+ * Method:    _getCurrent
+ * Signature: ()Lorg/postgresql/pljava/jdbc/Invocation;
+ */
+JNIEXPORT jobject JNICALL
+Java_org_postgresql_pljava_jdbc_Invocation__1getCurrent(JNIEnv* env, jclass cls)
+{
+	return currentCallContext->invocation;
+}
+
 /*
  * Class:     org_postgresql_pljava_jdbc_Invocation
  * Method:    _clearErrorCondition
