@@ -11,6 +11,7 @@
 
 #include "pljava/Backend.h"
 #include "pljava/Exception.h"
+#include "pljava/MemoryContext.h"
 #include "pljava/type/String.h"
 #include "pljava/type/ErrorData.h"
 
@@ -20,6 +21,9 @@ static jmethodID s_Class_getName;
 static jclass    s_Throwable_class;
 static jmethodID s_Throwable_getMessage;
 
+static jclass    s_IllegalArgumentException_class;
+static jmethodID s_IllegalArgumentException_init;
+
 static jclass    s_SQLException_class;
 static jmethodID s_SQLException_init;
 static jmethodID s_SQLException_getSQLState;
@@ -176,6 +180,27 @@ void Exception_throw(JNIEnv* env, int errCode, const char* errMessage, ...)
 	va_end(args);
 }
 
+void Exception_throwIllegalArgument(JNIEnv* env, const char* errMessage, ...)
+{
+    char buf[1024];
+	va_list args;
+	jstring message;
+	jobject ex;
+
+	va_start(args, errMessage);
+	vsnprintf(buf, sizeof(buf), errMessage, args);
+	ereport(DEBUG3, (errmsg(buf)));
+
+	message = String_createJavaStringFromNTS(env, buf);
+
+	ex = PgObject_newJavaObject(
+		env, s_IllegalArgumentException_class, s_IllegalArgumentException_init, message);
+
+	(*env)->DeleteLocalRef(env, message);
+	(*env)->Throw(env, ex);
+	va_end(args);
+}
+
 void Exception_throwSPI(JNIEnv* env, const char* function, int errCode)
 {
 #if (PGSQL_MAJOR_VER >= 8)
@@ -190,19 +215,25 @@ void Exception_throwSPI(JNIEnv* env, const char* function, int errCode)
 
 void Exception_throw_ERROR(JNIEnv* env, const char* funcName)
 {
+	jobject ed;
+	jobject ex;
+	ErrorData* errData;
 #if (PGSQL_MAJOR_VER >= 8)
-	ErrorData* errData = CopyErrorData();
+	MemoryContext_switchToUpperContext();
+	errData = CopyErrorData();
+	FlushErrorState();
 #else
 	StringInfoData buf;
-	ErrorData* errData = (ErrorData*)palloc(sizeof(ErrorData));
+	MemoryContext_switchToUpperContext();
+	errData = (ErrorData*)palloc(sizeof(ErrorData));
 	initStringInfo(&buf);
 	appendStringInfoString(&buf, "Error when calling: ");
 	appendStringInfoString(&buf, funcName);
 	errData->sqlerrcode = ERRCODE_INTERNAL_ERROR;
 	errData->message = buf.data;
 #endif
-	jobject ed = ErrorData_create(env, errData);
-	jobject ex = PgObject_newJavaObject(
+	ed = ErrorData_create(env, errData);
+	ex = PgObject_newJavaObject(
 		env, s_ServerException_class, s_ServerException_init, ed);
 	ereport(DEBUG3, (errcode(errData->sqlerrcode), errmsg(errData->message)));
 	currentCallContext->errorOccured = true;
@@ -225,6 +256,12 @@ Datum Exception_initialize(PG_FUNCTION_ARGS)
 	s_Throwable_getMessage = PgObject_getJavaMethod(env, s_Throwable_class,
 			"getMessage", "()Ljava/lang/String;");
 
+	s_IllegalArgumentException_class = (jclass)(*env)->NewGlobalRef(
+		env, PgObject_getJavaClass(env, "java/lang/IllegalArgumentException"));
+
+	s_IllegalArgumentException_init = PgObject_getJavaMethod(env, s_IllegalArgumentException_class,
+			"<init>", "(Ljava/lang/String;)V");
+
 	s_SQLException_class = (jclass)(*env)->NewGlobalRef(
 		env, PgObject_getJavaClass(env, "java/sql/SQLException"));
 
