@@ -7,11 +7,14 @@
  * @author Thomas Hallgren
  */
 #include "pljava/PgObject_priv.h"
+#include "pljava/backports.h"
 #include "pljava/Exception.h"
 #include "pljava/Backend.h"
 #include "pljava/HashMap.h"
 #include "pljava/MemoryContext.h"
 #include "pljava/type/Oid.h"
+#include "pljava/type/SingleRowWriter.h"
+#include "pljava/type/ResultSetProvider.h"
 #include "pljava/type/String.h"
 #include "pljava/type/TriggerData.h"
 
@@ -19,6 +22,7 @@
 #include <catalog/pg_namespace.h>
 #include <utils/builtins.h>
 #include <ctype.h>
+#include <funcapi.h>
 
 static jclass s_Loader_class;
 static jclass s_ClassLoader_class;
@@ -36,6 +40,12 @@ struct Function_
 	 */
 	bool      readOnly;
 
+	/*
+	 * True if the function is a multi-call function and hence, will
+	 * allocate a memory context of its own.
+	 */
+	bool      isMultiCall;
+
 	/*
 	 * True if the function returns a complex type.
 	 */
@@ -205,7 +215,7 @@ static void Function_parseParameters(Function self, Oid* dfltIds, const char* pa
 	}
 }
 
-static void Function_init(Function self, JNIEnv* env, Oid functionId, bool isTrigger)
+static void Function_init(Function self, JNIEnv* env, PG_FUNCTION_ARGS)
 {
 	const char* ip;
 	const char* paramDecl = 0;
@@ -225,10 +235,11 @@ static void Function_init(Function self, JNIEnv* env, Oid functionId, bool isTri
 	const char* ep;
 	const char* nameEp;
 	int32 len;
+	bool isResultSetProvider = false;
 
 	/* Obtain the tuple that corresponds to the function
 	 */
-	HeapTuple procTup = PgObject_getValidTuple(PROCOID, functionId, "function");
+	HeapTuple procTup = PgObject_getValidTuple(PROCOID, fcinfo->flinfo->fn_oid, "function");
 	Form_pg_proc procStruct = (Form_pg_proc)GETSTRUCT(procTup);
 
 	/* The user's function definition must be the fully
@@ -363,7 +374,7 @@ static void Function_init(Function self, JNIEnv* env, Oid functionId, bool isTri
 			errmsg("Extranious characters at end of method name '%s'", methodName)));
 
 	ctx = GetMemoryChunkContext(self);
-	if(isTrigger)
+	if(CALLED_AS_TRIGGER(fcinfo))
 	{
 		if(paramDecl != 0)
 			ereport(ERROR, (
@@ -382,55 +393,58 @@ static void Function_init(Function self, JNIEnv* env, Oid functionId, bool isTri
 	else
 	{
 		int top;
-		Oid retTypeId = procStruct->prorettype;
-		HeapTuple typeTup = PgObject_getValidTuple(TYPEOID, retTypeId, "type");
-		Form_pg_type pgType = (Form_pg_type)GETSTRUCT(typeTup);
 		Type complex = 0;
+		Oid retTypeId = InvalidOid;
+		TupleDesc retTuple = 0;
 
 		self->numParams = (int32)procStruct->pronargs;
+		self->isMultiCall = procStruct->proretset;
 
-		if(procStruct->proretset)
+		switch(get_call_result_type(fcinfo, &retTypeId, &retTuple))
 		{
-			/* The function returns a set. Check if the element type is a
-			 * complex or a scalar type.
-			 */
-			const char* javaName = (pgType->typtype == 'c')
-				? "org.postgresql.pljava.ResultSetProvider"
-				: "java.util.Iterator";
+			case TYPEFUNC_SCALAR:
+				if(self->isMultiCall)
+					self->returnType = Type_fromJavaType(retTypeId, "java.util.Iterator");
+				else
+				{
+					HeapTuple typeTup = PgObject_getValidTuple(TYPEOID, retTypeId, "type");
+					Form_pg_type pgType = (Form_pg_type)GETSTRUCT(typeTup);
+					self->returnType = Type_fromPgType(retTypeId, pgType);
+					ReleaseSysCache(typeTup);
+				}
+				break;
 
-			/* Obtain the provider type for the set type.
-			 */
-			self->returnType = Type_fromJavaType(retTypeId, javaName);
-		}
-		else
-		{
-			/*
-			 * Retreive standard string conversion from the postgres
-			 * type catalog.
-			 */
-			if(pgType->typtype == 'c')
-			{
-				/* Complex functions uses an updateable ResultSet
-				 * as the last argument and returns boolean to indicate
-				 * whether or not this set has been filled in.
-				 */
-				complex = Type_fromJavaType(
-					retTypeId,
-					"org.postgresql.pljava.jdbc.SingleRowWriter");
-				self->returnType = Type_fromOid(BOOLOID);
-				self->numParams++;
-				self->returnComplex = true;
-			}
-			else
-				self->returnType = Type_fromPgType(retTypeId, pgType);
-		}
-		ReleaseSysCache(typeTup);
+			case TYPEFUNC_COMPOSITE:
+			case TYPEFUNC_RECORD:
+				if(self->isMultiCall)
+				{
+					isResultSetProvider = true;
+					self->returnType = ResultSetProvider_createType(retTypeId, retTuple);
+				}
+				else
+				{
+					self->numParams++;
+					self->returnComplex = true;
+					self->returnType = Type_fromOid(BOOLOID);
+					complex = SingleRowWriter_createType(retTypeId, retTuple);
+				}
+				break;
 
+			case TYPEFUNC_OTHER:
+				ereport(ERROR, (
+					errcode(ERRCODE_SYNTAX_ERROR),
+					errmsg("PL/Java functions cannot return type %s",
+						format_type_be(procStruct->prorettype))));
+		}
 		top = self->numParams;
 		if(top > 0)
 		{
 			int idx;
+#if (PGSQL_MAJOR_VER < 8 || (PGSQL_MAJOR_VER == 8 && PGSQL_MINOR_VER == 0))
 			Oid* typeIds = procStruct->proargtypes;
+#else
+			Oid* typeIds = procStruct->proargtypes.values;
+#endif
 			self->paramTypes = (Type*)MemoryContextAlloc(ctx, top * sizeof(Type));
 
 			if(complex != 0)
@@ -484,19 +498,37 @@ static void Function_init(Function self, JNIEnv* env, Oid functionId, bool isTri
 	if(self->method == 0)
 	{
 		char* origSign = sign.data;
+		Type altType = 0;
+		Type realRetType = self->returnType;
+
 		elog(DEBUG1, "Method %s.%s %s not found", className, methodName, origSign);
+
 		if(Type_isPrimitive(self->returnType))
 		{
 			/*
-			 * There's one valid reason for not finding the method, namely
-			 * if the return type used in the signature is a primitive and
+			 * One valid reason for not finding the method is when
+			 * the return type used in the signature is a primitive and
 			 * the true return type of the method is the object class that
 			 * corresponds to that primitive.
 			 */
-			Type objType = Type_getObjectType(self->returnType);
+			altType = Type_getObjectType(self->returnType);
+			realRetType = altType;
+		}
+		else if(isResultSetProvider)
+		{
+			/*
+			 * Another reason might be that we expected a ResultSetProvider
+			 * but the implementation returns a ResultSetHandle that needs to be
+			 * wrapped. The wrapping is internal so we retain the original
+			 * return type anyway.
+			 */
+			altType = Type_fromJavaType(InvalidOid, "org.postgresql.pljava.ResultSetHandle");
+		}
 
+		if(altType != 0)
+		{
 			(*env)->ExceptionClear(env);
-			Function_buildSignature(self, &sign, objType);
+			Function_buildSignature(self, &sign, altType);
 	
 			elog(DEBUG1, "Obtaining method %s.%s %s", className, methodName, sign.data);
 			isCallingJava = true;
@@ -504,34 +536,35 @@ static void Function_init(Function self, JNIEnv* env, Oid functionId, bool isTri
 							env, self->clazz, methodName, sign.data);
 			isCallingJava = saveIcj;
 	
-			if(self->method == 0)
-				PgObject_throwMemberError(env, self->clazz, methodName, origSign, true, true);
-	
-			pfree(origSign);
-			self->returnType = objType;
+			if(self->method != 0)
+				self->returnType = realRetType;
 		}
-		else
+		if(self->method == 0)
 			PgObject_throwMemberError(env, self->clazz, methodName, origSign, true, true);
+
+		if(sign.data != origSign)
+			pfree(origSign);
 	}
 	pfree(sign.data);
 	pfree(className);
 }
 
-static Function Function_create(JNIEnv* env, Oid functionId, bool isTrigger)
+static Function Function_create(JNIEnv* env, PG_FUNCTION_ARGS)
 {
 	Function self = (Function)PgObjectClass_allocInstance(s_FunctionClass, TopMemoryContext);
-	Function_init(self, env, functionId, isTrigger);
+	Function_init(self, env, fcinfo);
 	return self;
 }
 
-Function Function_getFunction(JNIEnv* env, Oid functionId, bool isTrigger)
+Function Function_getFunction(JNIEnv* env, PG_FUNCTION_ARGS)
 {
-	Function func = (Function)HashMap_getByOid(s_funcMap, functionId);
+	Oid funcOid = fcinfo->flinfo->fn_oid;
+	Function func = (Function)HashMap_getByOid(s_funcMap, funcOid);
 	if(func == 0)
 	{
 		PgObject old;
-		func = Function_create(env, functionId, isTrigger);
-		old = HashMap_putByOid(s_funcMap, functionId, func);
+		func = Function_create(env, fcinfo);
+		old = HashMap_putByOid(s_funcMap, funcOid, func);
 		if(old != 0)
 		{
 			/* Can happen in a multithreaded environment. Extremely
@@ -555,9 +588,16 @@ Datum Function_invoke(Function self, JNIEnv* env, PG_FUNCTION_ARGS)
 	{
 		int32   idx;
 		Type    invokerType;
-		jvalue* args  = (jvalue*)palloc(top * sizeof(jvalue));
+		jvalue* args;
 		Type*   types = self->paramTypes;
 
+		/* a class loader or other mechanism might have connected already. This
+		 * connection must be dropped since its parent context is wrong.
+		 */
+		if(self->isMultiCall && SRF_IS_FIRSTCALL())
+			Backend_assertDisconnect();
+
+		args  = (jvalue*)palloc(top * sizeof(jvalue));
 		if(self->returnComplex)
 		{
 			--top; /* Last argument is not present in fcinfo */
