@@ -7,6 +7,7 @@
  * @author Thomas Hallgren
  */
 #include "org_postgresql_pljava_internal_SPI.h"
+#include "pljava/backports.h"
 #include "pljava/SPI.h"
 #include "pljava/Backend.h"
 #include "pljava/Exception.h"
@@ -68,6 +69,7 @@ Java_org_postgresql_pljava_internal_SPI__1exec(JNIEnv* env, jclass cls, jstring
 	Backend_pushJavaFrame(env);
 	PG_TRY();
 	{
+		Backend_assertConnect();
 		result = (jint)SPI_exec(command, (int)count);
 		if(result < 0)
 			Exception_throwSPI(env, "exec", result);
@@ -118,51 +120,6 @@ Java_org_postgresql_pljava_internal_SPI__1getTupTable(JNIEnv* env, jclass cls)
 	return SPITupleTable_create(env, SPI_tuptable);
 }
 
-#include <executor/spi_priv.h> /* Needed to get to the argtypes of the plan */
-
-#if (PGSQL_MAJOR_VER == 7 && PGSQL_MINOR_VER < 5)
-
-Oid SPI_getargtypeid(void* plan, int argIndex)
-{
-	if (plan == NULL || argIndex < 0 || argIndex >= ((_SPI_plan*)plan)->nargs)
-	{
-		SPI_result = SPI_ERROR_ARGUMENT;
-		return InvalidOid;
-	}
-	return ((_SPI_plan*)plan)->argtypes[argIndex];
-}
-
-int SPI_getargcount(void* plan)
-{
-	if (plan == NULL)
-	{
-		SPI_result = SPI_ERROR_ARGUMENT;
-		return -1;
-	}
-	return ((_SPI_plan*)plan)->nargs;
-}
-
-bool SPI_is_cursor_plan(void* plan)
-{
-	List* qtlist;
-	_SPI_plan* spiplan = (_SPI_plan*)plan;
-	if (spiplan == NULL)
-	{
-		SPI_result = SPI_ERROR_ARGUMENT;
-		return false;
-	}
-
-	qtlist = spiplan->qtlist;
-	if(length(spiplan->ptlist) == 1 && length(qtlist) == 1)
-	{
-		Query* queryTree = (Query*)lfirst((List*)lfirst(qtlist));
-		if(queryTree->commandType == CMD_SELECT && queryTree->into == NULL)
-			return true;
-	}
-	return false;
-}
-#endif
-
 #if (PGSQL_MAJOR_VER >= 8)
 static void assertXid(SubTransactionId xid)
 {
@@ -183,6 +140,7 @@ Savepoint* SPI_setSavepoint(const char* name)
 	 * or rolled back even if the creator forgets about it.
 	 */
 	Savepoint* sp = (Savepoint*)palloc(sizeof(Savepoint) + strlen(name));
+	Backend_assertConnect();
 	BeginInternalSubTransaction((char*)name);
 	sp->nestingLevel = GetCurrentTransactionNestLevel();
 	sp->xid = GetCurrentSubTransactionId();
@@ -213,6 +171,7 @@ void SPI_rollbackSavepoint(Savepoint* sp)
 		assertXid(sp->xid);
 		RollbackAndReleaseCurrentSubTransaction();
 	}
+	SPI_restore_connection();
 	pfree(sp);
 }
 #endif
