@@ -17,6 +17,7 @@
 #include "pljava/type/String.h"
 #include "pljava/type/Tuple.h"
 #include "pljava/type/TupleDesc.h"
+#include "pljava/type/Oid.h"
 
 static Type      s_TupleDesc;
 static TypeClass s_TupleDescClass;
@@ -54,7 +55,23 @@ TupleDesc TupleDesc_forOid(Oid oid)
 	}
 	return tupleDesc;
 }
-	
+
+Type TupleDesc_getColumnType(JNIEnv* env, TupleDesc tupleDesc, int index)
+{
+	Type type;
+	Oid typeId = SPI_gettypeid(tupleDesc, index);
+	if(!OidIsValid(typeId))
+	{
+		Exception_throw(env,
+			ERRCODE_INVALID_DESCRIPTOR_INDEX,
+			"Invalid attribute index \"%d\"", (int)index);
+		type = 0;
+	}
+	else
+		type = Type_objectTypeFromOid(typeId);
+	return type;
+}
+
 static jvalue _TupleDesc_coerceDatum(Type self, JNIEnv* env, Datum arg)
 {
 	jvalue result;
@@ -94,6 +111,11 @@ Datum TupleDesc_initialize(PG_FUNCTION_ARGS)
 		"()I",
 		Java_org_postgresql_pljava_internal_TupleDesc__1size
 		},
+		{
+		"_getOid",
+		"(I)Lorg/postgresql/pljava/internal/Oid;",
+		Java_org_postgresql_pljava_internal_TupleDesc__1getOid
+		},
 		{ 0, 0, 0 }};
 
 	JNIEnv* env = (JNIEnv*)PG_GETARG_POINTER(0);
@@ -264,3 +286,41 @@ Java_org_postgresql_pljava_internal_TupleDesc__1size(JNIEnv* env, jobject _this)
 		return 0;
 	return (jint)self->natts;
 }
+
+/*
+ * Class:     org_postgresql_pljava_internal_TupleDesc
+ * Method:    _getOid
+ * Signature: (I)Lorg/postgresql/pljava/internal/Oid;
+ */
+JNIEXPORT jobject JNICALL
+Java_org_postgresql_pljava_internal_TupleDesc__1getOid(JNIEnv* env, jobject _this, jint index)
+{
+	TupleDesc self;
+	jobject result = 0;
+	PLJAVA_ENTRY_FENCE(0)
+
+	self = (TupleDesc)NativeStruct_getStruct(env, _this);
+	if(self != 0)
+	{
+		PG_TRY();
+		{
+			Oid typeId = SPI_gettypeid(self, (int)index);
+			if(!OidIsValid(typeId))
+			{
+				Exception_throw(env,
+					ERRCODE_INVALID_DESCRIPTOR_INDEX,
+					"Invalid attribute index \"%d\"", (int)index);
+			}
+			else
+			{
+				result = Oid_create(env, typeId);
+			}
+		}
+		PG_CATCH();
+		{
+			Exception_throw_ERROR(env, "SPI_gettypeid");
+		}
+		PG_END_TRY();
+	}
+	return result;
+}
