@@ -7,12 +7,14 @@
 package org.postgresql.pljava.example;
 
 import java.sql.Connection;
+import java.sql.Date;
 import java.sql.DriverManager;
 import java.sql.PreparedStatement;
 import java.sql.ResultSet;
 import java.sql.SQLException;
 import java.sql.Savepoint;
 import java.sql.Statement;
+import java.sql.Time;
 import java.util.logging.Logger;
 
 /**
@@ -42,7 +44,7 @@ public class SPIActions
 
 		// Create an anonymous savepoint.
 		//
-		log("Attempting to set an anonymous savepoint, expect to see an UnsupportedOperationException for versions priort to 8.0");
+		log("Attempting to set an anonymous savepoint, expect to see an UnsupportedOperationException for versions prior to 8.0");
 		try
 		{
 			Savepoint sp = conn.setSavepoint();
@@ -68,6 +70,52 @@ public class SPIActions
 		throw new SQLException("SAVEPOINT through SQL succeeded. That's bad news!");
 	}
 
+	public static int testTransactionRecovery()
+	throws SQLException
+	{
+		Connection conn = DriverManager.getConnection("jdbc:default:connection");
+
+		// Create an anonymous savepoint.
+		//
+		log("Attempting to set an anonymous savepoint, expect to see an UnsupportedOperationException for versions prior to 8.0");
+		try
+		{
+    		Statement stmt = conn.createStatement();
+			Savepoint sp = conn.setSavepoint();
+
+            try
+			{
+				log("Attempting to execute a statement with a syntax error");
+				stmt.execute("THIS MUST BE A SYNTAX ERROR");
+			}
+			catch(SQLException e)
+			{
+				log("It failed. Let's try to recover " +
+                    "by rolling back to anonymous savepoint");
+				conn.rollback(sp);
+				log("Rolled back.");
+				log("Now let's try to execute a correct statement.");
+                ResultSet rs = stmt.executeQuery("SELECT 'OK'");
+                while (rs.next())
+                {
+                    log("Expected: OK; Retrieved: " + rs.getString(1));
+                }
+                rs.close();
+                stmt.close();
+                
+				return 1;
+			}
+		}
+		catch(UnsupportedOperationException e)
+		{
+			log(e.getMessage());
+			return 0;
+		}
+		
+        //Should never get here
+        return -1;
+	}
+
 	public static int transferPeopleWithSalary(int salary)
 	throws SQLException
 	{
@@ -84,7 +132,7 @@ public class SPIActions
 			log(stmt);
 			select = conn.prepareStatement(stmt);
 
-			stmt = "INSERT INTO employees2(id, name, salary) VALUES (?, ?, ?)";
+			stmt = "INSERT INTO employees2(id, name, salary, transferDay, transferTime) VALUES (?, ?, ?, ?, ?)";
 			log(stmt);
 			insert = conn.prepareStatement(stmt);
 
@@ -108,9 +156,12 @@ public class SPIActions
 				insert.setInt(1, id);
 				insert.setString(2, name);
 				insert.setInt(3, empSal);
+				long now = System.currentTimeMillis();
+				insert.setDate(4, new Date(now));
+				insert.setTime(5, new Time(now));
 				int nRows = insert.executeUpdate();
 				log("Insert processed " + nRows + " rows");
-				
+
 				delete.setInt(1, id);
 				nRows = delete.executeUpdate();
 				log("Delete processed " + nRows + " rows");
@@ -163,4 +214,52 @@ public class SPIActions
 			conn.close();
 		}
 	}
+
+	public static String getDateAsString()
+	throws SQLException
+	{
+		ResultSet rs = null;
+		Statement stmt = null;
+		Connection conn = DriverManager.getConnection("jdbc:default:connection");
+		try
+		{
+			stmt = conn.createStatement();
+			rs = stmt.executeQuery("SELECT CURRENT_DATE");
+			if(rs.next())
+				return rs.getDate(1).toString();
+			return "Date could not be retrieved";
+		}
+		finally
+		{
+			if(rs != null)
+				rs.close();
+			if(stmt != null)
+				stmt.close();
+			conn.close();
+		}
+	}
+
+	public static String getTimeAsString()
+	throws SQLException
+	{
+		ResultSet rs = null;
+		Statement stmt = null;
+		Connection conn = DriverManager.getConnection("jdbc:default:connection");
+		try
+		{
+			stmt = conn.createStatement();
+			rs = stmt.executeQuery("SELECT CURRENT_TIME");
+			if(rs.next())
+				return rs.getTime(1).toString();
+			return "Time could not be retrieved";
+		}
+		finally
+		{
+			if(rs != null)
+				rs.close();
+			if(stmt != null)
+				stmt.close();
+			conn.close();
+		}
+	}
 }
