@@ -25,9 +25,14 @@ import java.sql.Time;
 import java.sql.Timestamp;
 import java.sql.Types;
 import java.util.BitSet;
+import java.util.Calendar;
 import java.util.HashMap;
 import java.util.Map;
+import java.util.regex.Pattern;
+import java.util.regex.Matcher;
+import java.util.regex.PatternSyntaxException;
 
+import org.postgresql.pljava.internal.Oid;
 
 /**
  * @author Thomas Hallgren
@@ -35,6 +40,7 @@ import java.util.Map;
 public class SPIConnection implements Connection
 {
 	private static final HashMap s_sqlType2Class = new HashMap(30);
+    private int[] VERSION_NUMBER = null; //version number
 
 	static
 	{
@@ -225,7 +231,7 @@ public class SPIConnection implements Connection
 	public DatabaseMetaData getMetaData()
 	throws SQLException
 	{
-		throw new UnsupportedFeatureException("Connection.getMetaData");
+		return new SPIDatabaseMetaData(this);
 	}
 
 	/**
@@ -575,4 +581,286 @@ public class SPIConnection implements Connection
 		if(invocation.getSavepoint() == sp)
 			invocation.setSavepoint(null);
 	}
-}
+
+    public int[] getVersionNumber() throws SQLException
+    {
+        if (VERSION_NUMBER != null)
+        	return VERSION_NUMBER;
+
+        ResultSet rs = createStatement().executeQuery(
+            "SELECT version()");
+
+        try
+        {
+            if (!rs.next())
+                throw new SQLException(
+                "Cannot retrieve product version number");
+
+            String ver = rs.getString(1);
+            Pattern p = Pattern.compile(
+                "^PostgreSQL\\s+(\\d+)\\.(\\d+)\\.(\\d+)\\s+.*");
+            Matcher m = p.matcher(ver);
+            if (m.matches() && m.groupCount() == 3)
+            {
+            	VERSION_NUMBER = new int[3];
+                VERSION_NUMBER[0] = Integer.parseInt(m.group(1));
+                VERSION_NUMBER[1] = Integer.parseInt(m.group(2));
+                VERSION_NUMBER[2] = Integer.parseInt(m.group(3));
+                return VERSION_NUMBER;
+            }
+            throw new SQLException(
+                "Unexpected product version string format: " +
+                ver);
+        }
+        catch (PatternSyntaxException e)
+        {
+            throw new SQLException(
+                "Error in product version string parsing: " +
+                e.getMessage());
+        }
+        finally
+        {
+            rs.close();
+        }
+    }
+
+    /*
+     * This implemetation uses the jdbc3Types array to support the jdbc3
+     * datatypes.  Basically jdbc2 and jdbc3 are the same, except that
+     * jdbc3 adds some
+     */
+    public int getSQLType(String pgTypeName)
+    {
+        if (pgTypeName == null)
+            return Types.OTHER;
+
+        for (int i = 0;i < JDBC3_TYPE_NAMES.length;i++)
+            if (pgTypeName.equals(JDBC3_TYPE_NAMES[i]))
+                return JDBC_TYPE_NUMBERS[i];
+
+        return Types.OTHER;
+    }
+
+    /*
+     * This returns the java.sql.Types type for a PG type oid
+     *
+     * @param oid PostgreSQL type oid
+     * @return the java.sql.Types type
+     * @exception SQLException if a database access error occurs
+     */
+    public int getSQLType(Oid oid) throws SQLException
+    {
+        return getSQLType(getPGType(oid));
+    }
+ 
+    public String getPGType(Oid oid) throws SQLException
+    {
+        String typeName = null;
+        PreparedStatement query = null;
+        ResultSet rs = null;
+
+        try
+        {
+            query = prepareStatement("SELECT typname FROM pg_catalog.pg_type WHERE oid=?");
+            query.setObject(1, oid);
+            rs = query.executeQuery();
+
+            if (rs.next())
+            {
+                typeName = rs.getString(1);
+            }
+            else
+            {
+                throw new SQLException("Cannot find PG type with oid=" + oid);
+            }
+        }
+        finally
+        {
+            if (query != null)
+            {
+                query.close();
+            }
+        }
+
+        return typeName;
+    }
+
+	static Object basicCoersion(Class cls, Object value)
+	throws SQLException
+	{
+		if(value == null || cls.isInstance(value))
+			return value;
+
+		if(cls == String.class)
+		{
+			if(value instanceof Number
+			|| value instanceof Boolean
+			|| value instanceof Timestamp
+			|| value instanceof Date
+			|| value instanceof Time)
+				return value.toString();
+		}
+		throw new SQLException("Cannot derive a value of class " +
+				cls.getName() + " from an object of class " + value.getClass().getName());
+	}
+
+	static Number basicNumericCoersion(Class cls, Object value)
+	throws SQLException
+	{
+		if(value == null || value instanceof Number)
+			return (Number)value;
+
+		if(cls == int.class  || cls == long.class || cls == short.class || cls == byte.class)
+		{
+			if(value instanceof String)
+				return Long.valueOf((String)value);
+
+			if(value instanceof Boolean)
+				return new Long(((Boolean)value).booleanValue() ? 1 : 0);
+		}
+		else if(cls == BigDecimal.class)
+		{
+			if(value instanceof String)
+				return new BigDecimal((String)value);
+
+			if(value instanceof Boolean)
+				return new BigDecimal(((Boolean)value).booleanValue() ? 1 : 0);
+		}
+		if(cls == double.class  || cls == float.class)
+		{
+			if(value instanceof String)
+				return Double.valueOf((String)value);
+
+			if(value instanceof Boolean)
+				return new Double(((Boolean)value).booleanValue() ? 1 : 0);
+		}
+		throw new SQLException("Cannot derive a Number from an object of class " + value.getClass().getName());
+	}
+
+	static Object basicCalendricalCoersion(Class cls, Object value, Calendar cal)
+	throws SQLException
+	{
+		if(value == null)
+			return value;
+
+		if(cls.isInstance(value))
+			return value;
+
+		if(cls == Timestamp.class)
+		{
+			if(value instanceof Date)
+			{
+				cal.setTime((Date)value);
+				cal.set(Calendar.HOUR_OF_DAY, 0);
+				cal.set(Calendar.MINUTE, 0);
+				cal.set(Calendar.SECOND, 0);
+				cal.set(Calendar.MILLISECOND, 0);
+				return new Timestamp(cal.getTimeInMillis());
+			}
+			else if(value instanceof Time)
+			{
+				cal.setTime((Date)value);
+				cal.set(1970, 0, 1);
+				return new Timestamp(cal.getTimeInMillis());
+			}
+			else if(value instanceof String)
+			{
+				return Timestamp.valueOf((String)value);
+			}
+		}
+		else if(cls == Date.class)
+		{
+			if(value instanceof Timestamp)
+			{
+				Timestamp ts = (Timestamp)value;
+				cal.setTime(ts);
+				cal.set(Calendar.HOUR_OF_DAY, 0);
+				cal.set(Calendar.MINUTE, 0);
+				cal.set(Calendar.SECOND, 0);
+				cal.set(Calendar.MILLISECOND, 0);
+				return new Date(cal.getTimeInMillis());
+			}
+			else if(value instanceof String)
+			{
+				return Date.valueOf((String)value);
+			}
+		}
+		else if(cls == Time.class)
+		{
+			if(value instanceof Timestamp)
+			{
+				Timestamp ts = (Timestamp)value;
+				cal.setTime(ts);
+				cal.set(1970, 0, 1);
+				return new Time(cal.getTimeInMillis());
+			}
+			else if(value instanceof String)
+			{
+				return Time.valueOf((String)value);
+			}
+		}
+		throw new SQLException("Cannot derive a value of class " +
+			cls.getName() + " from an object of class " + value.getClass().getName());
+	}
+
+    /*
+     * This table holds the org.postgresql names for the types supported.
+     * Any types that map to Types.OTHER (eg POINT) don't go into this table.
+     * They default automatically to Types.OTHER
+     *
+     * Note: This must be in the same order as below.
+     *
+     * Tip: keep these grouped together by the Types. value
+     */
+    public static final String JDBC3_TYPE_NAMES[] = {
+                "int2",
+                "int4", "oid",
+                "int8",
+                "cash", "money",
+                "numeric",
+                "float4",
+                "float8",
+                "bpchar", "char", "char2", "char4", "char8", "char16",
+                "varchar", "text", "name", "filename",
+                "bytea",
+                "bool",
+                "bit",
+                "date",
+                "time", "timetz",
+                "abstime", "timestamp", "timestamptz",
+                "_bool", "_char", "_int2", "_int4", "_text",
+                "_oid", "_varchar", "_int8", "_float4", "_float8",
+                "_abstime", "_date", "_time", "_timestamp", "_numeric",
+                "_bytea"
+            };
+
+    /*
+     * This table holds the JDBC type for each entry above.
+     *
+     * Note: This must be in the same order as above
+     *
+     * Tip: keep these grouped together by the Types. value
+     */
+    public static final int JDBC_TYPE_NUMBERS[] =
+    		{
+                Types.SMALLINT,
+                Types.INTEGER, Types.INTEGER,
+                Types.BIGINT,
+                Types.DOUBLE, Types.DOUBLE,
+                Types.NUMERIC,
+                Types.REAL,
+                Types.DOUBLE,
+                Types.CHAR, Types.CHAR, Types.CHAR, Types.CHAR, Types.CHAR, Types.CHAR,
+                Types.VARCHAR, Types.VARCHAR, Types.VARCHAR, Types.VARCHAR,
+                Types.BINARY,
+                Types.BIT,
+                Types.BIT,
+                Types.DATE,
+                Types.TIME, Types.TIME,
+                Types.TIMESTAMP, Types.TIMESTAMP, Types.TIMESTAMP,
+                Types.ARRAY, Types.ARRAY, Types.ARRAY, Types.ARRAY, Types.ARRAY,
+                Types.ARRAY, Types.ARRAY, Types.ARRAY, Types.ARRAY, Types.ARRAY,
+                Types.ARRAY, Types.ARRAY, Types.ARRAY, Types.ARRAY, Types.ARRAY,
+                Types.ARRAY
+            };
+ }
