@@ -13,6 +13,7 @@ import java.sql.ResultSet;
 import java.sql.Statement;
 import java.sql.SQLException;
 import java.util.ArrayList;
+import java.util.Arrays;
 import java.io.PrintStream;
 
 /**
@@ -64,8 +65,7 @@ public class Tester
 	public static void printUsage()
 	{
 		PrintStream out = System.err;
-		out.println("usage: java -jar deploy.jar");
-		out.println("    {-install | -uninstall | -reinstall}");
+		out.println("usage: java org.postgresql.pljava.test.Tester");
 		out.println("    [ -host <hostName>     ]    # default is localhost");
 		out.println("    [ -port <portNubmer>   ]    # default is blank");
 		out.println("    [ -database <database> ]    # default is name of current user");
@@ -205,7 +205,12 @@ public class Tester
 				t.testCurrentDir();
 				t.testUsingProperties();
 				t.testUsingScalarProperties();
+				t.testUsingResultSetProperties();
 				t.testSavepointSanity();
+				t.testTrustedSecurity();
+				t.testBinaryColumns();
+				t.testDatabaseMetaData();
+				t.testResultSet();
 			}
 			t.close();
 		}
@@ -318,6 +323,21 @@ public class Tester
 		rs.close();
 	}
 
+	public void testUsingResultSetProperties()
+	throws SQLException
+	{
+		System.out.println("*** testUsingResultSetProperties()");
+		Statement stmt = m_connection.createStatement();
+		ResultSet rs = stmt.executeQuery("SELECT name, value FROM resultSetPropertyExample()");
+		while(rs.next())
+		{
+			String name = rs.getString(1);
+			String value = rs.getString(2);
+			System.out.println("Name = \"" + name + "\", value = \"" + value + "\"");
+		}
+		rs.close();
+	}
+
 	public void testUsingScalarProperties()
 	throws SQLException
 	{
@@ -329,6 +349,23 @@ public class Tester
 		rs.close();
 	}
 
+	public void testBinaryColumns()
+	throws SQLException
+	{
+		System.out.println("*** testBinaryColumns()");
+
+		Statement stmt = m_connection.createStatement();
+		ResultSet rs = stmt.executeQuery("SELECT * FROM binaryColumnTest()");
+		while(rs.next())
+		{
+			byte[] b1 = rs.getBytes(1);
+			byte[] b2 = rs.getBytes(2);
+			if(!Arrays.equals(b1, b2))
+				throw new SQLException("binary columns differ");
+		}
+		rs.close();
+	}
+
 	public void testCallInCall()
 	throws SQLException
 	{
@@ -507,4 +544,248 @@ public class Tester
 		rs.close();
 		stmt.close();
 	}
+
+	public void testTrustedSecurity()
+	throws SQLException
+	{
+		System.out.println("*** testTrustedSecurity()");
+		Statement stmt = m_connection.createStatement();
+		ResultSet rs = null;
+		
+		tryCreateTempFile(true);
+
+		boolean funcCreated = false;
+		try
+		{
+			rs = stmt.executeQuery("SHOW IS_SUPERUSER");
+			if(!(rs.next() && rs.getString(1).equals("on")))
+			{
+				System.out.println("Tester is not superuser so tests on untrusted language cannot be performed");
+				return;
+			}
+
+			// Try the same java method again, this time using language javaU
+			//
+			System.out.println("*** testUntrustedSecurity()");
+			stmt.execute(
+				"CREATE OR REPLACE FUNCTION javatest.create_temp_file_untrusted()" +
+				"  RETURNS varchar" +
+				"  AS 'org.postgresql.pljava.example.Security.createTempFile'" +
+				"  LANGUAGE javaU");
+			
+			tryCreateTempFile(false);
+		}
+		finally
+		{
+			if(rs != null)
+			{
+				try { rs.close(); } catch(SQLException e) {}
+				rs = null;
+			}
+			if(funcCreated)
+				stmt.execute("DROP FUNCTION javatest.create_temp_file_untrusted()");
+			stmt.close();
+		}
+	}
+	
+	private void tryCreateTempFile(boolean trusted)
+	throws SQLException
+	{
+		Statement stmt = m_connection.createStatement();
+		ResultSet rs = null;
+
+		try
+		{
+			if(trusted)
+				rs = stmt.executeQuery("SELECT create_temp_file_trusted()");
+			else
+				rs = stmt.executeQuery("SELECT create_temp_file_untrusted()");
+
+			if(!rs.next())
+				System.out.println("Unable to position ResultSet");
+			else
+				System.out.println("Name of created temp file = " + rs.getString(1));
+
+			if(trusted)
+				throw new RuntimeException("ERROR: Tempfile creation succeded although language is trusted!");
+		}
+		catch(SQLException e)
+		{
+			if(!trusted)
+				throw e;
+			System.out.println("OK, creation of temp file was *unsuccessful* as it should be");
+		}
+		finally
+		{
+			if(rs != null)
+			{
+				try { rs.close(); } catch(SQLException e) {}
+				rs = null;
+			}
+			stmt.close();
+		}
+	}
+
+	public void testDatabaseMetaData()
+	throws SQLException
+	{
+		Statement stmt = m_connection.createStatement();
+		ResultSet rs = null;
+
+		try
+		{
+			System.out.println("*** DatabaseMetaData 'String' functions:");
+			rs = stmt.executeQuery("SELECT * FROM javatest.getMetaDataStrings()");
+			while(rs.next())
+			{
+				String methodName = rs.getString(1);
+				String methodResult = rs.getString(2);
+				System.out.println(
+						"Method = \"" + methodName +
+						"\", result = \"" + methodResult + "\"");
+			}
+			rs.close();
+
+			System.out.println("*** DatabaseMetaData 'boolean' functions:");
+			rs = stmt.executeQuery("SELECT * FROM javatest.getMetaDataBooleans()");
+			while(rs.next())
+			{
+				String methodName = rs.getString(1);
+				boolean methodResult = rs.getBoolean(2);
+				System.out.println(
+						"Method = \"" + methodName +
+						"\", result = \"" + methodResult + "\"");
+			}
+			rs.close();
+
+			System.out.println("*** DatabaseMetaData 'int' functions:");
+			rs = stmt.executeQuery("SELECT * FROM javatest.getMetaDataInts()");
+			while(rs.next())
+			{
+				String methodName = rs.getString(1);
+				int methodResult = rs.getInt(2);
+				System.out.println(
+						"Method = \"" + methodName +
+						"\", result = \"" + methodResult + "\"");
+			}
+			rs.close();
+
+			executeMetaDataFunction(stmt,
+				"getAttributes((String)null,\"javatest\",\"%\",\"%\")");
+			executeMetaDataFunction(stmt,
+				"getBestRowIdentifier((String)null,\"sqlj\",\"jar_repository\",0,FALSE)");
+			executeMetaDataFunction(stmt,
+				"getCatalogs()");
+			executeMetaDataFunction(stmt,
+				"getColumnPrivileges((String)null,\"sqlj\",\"jar_repository\",\"jarid\")");
+			executeMetaDataFunction(stmt,
+				"getColumns((String)null,\"sqlj\",\"jar_repository\",\"%\")");
+			executeMetaDataFunction(stmt,
+				"getCrossReference((String)null,\"sqlj\",\"jar_repository\",(String)null,\"sqlj\",\"jar_entry\")");
+			executeMetaDataFunction(stmt,
+				"getExportedKeys((String)null,\"sqlj\",\"jar_repository\")");
+			executeMetaDataFunction(stmt,
+				"getImportedKeys((String)null,\"sqlj\",\"jar_repository\")");
+			executeMetaDataFunction(stmt,
+				"getIndexInfo((String)null,\"sqlj\",\"jar_repository\",TRUE,FALSE)");
+			executeMetaDataFunction(stmt,
+				"getPrimaryKeys((String)null,\"sqlj\",\"jar_repository\")");
+			executeMetaDataFunction(stmt,
+				"getProcedureColumns((String)null,\"sqlj\",\"install_jar\",(String)null)");
+			executeMetaDataFunction(stmt,
+				"getProcedures((String)null,\"sqlj\",\"%\")");
+			executeMetaDataFunction(stmt,
+				"getSchemas()");
+			executeMetaDataFunction(stmt,
+				"getSuperTables((String)null,\"sqlj\",\"jar_repository\")");
+			executeMetaDataFunction(stmt,
+				"getSuperTypes((String)null,\"sqlj\",\"%\")");
+			executeMetaDataFunction(stmt,
+				"getTablePrivileges((String)null,\"sqlj\",\"jar_repository\")");
+			executeMetaDataFunction(stmt,
+				"getTables((String)null,\"sqlj\",\"jar%\",{\"TABLE\"})");
+			executeMetaDataFunction(stmt,
+				"getTableTypes()");
+			executeMetaDataFunction(stmt,
+				"getTypeInfo()");
+			executeMetaDataFunction(stmt,
+				"getUDTs((String)null,\"sqlj\",\"%\",(int[])null)");
+			executeMetaDataFunction(stmt,
+				"getVersionColumns((String)null,\"sqlj\",\"jar_repository\")");
+		}
+		finally
+		{
+			if(rs != null)
+			{
+				try { rs.close(); } catch(SQLException e) {}
+				rs = null;
+			}
+		}
+	}
+
+	private void executeMetaDataFunction(Statement stmt, String functionCall)
+	throws SQLException
+	{
+		ResultSet rs = null;
+		try
+		{
+			System.out.println("*** " + functionCall + ":");
+			rs = stmt.executeQuery("SELECT * FROM javatest.callMetaDataMethod('" +
+								   functionCall + "')");
+			while(rs.next())
+			{
+				System.out.println(rs.getString(1));
+			}
+			rs.close();
+
+		}
+		catch (Exception e)
+		{
+			System.out.println("  Failed: " + e.getMessage());
+		}
+		finally
+		{
+			if(rs != null)
+			{
+				try { rs.close(); } catch(SQLException e) {}
+				rs = null;
+			}
+		}
+	}
+
+	public void testResultSet()
+	throws SQLException
+	{
+        String sql;
+		Statement stmt = m_connection.createStatement();
+		ResultSet rs = null;
+
+		try
+		{
+			System.out.println("*** ResultSet test:");
+            sql =
+                "SELECT * FROM javatest.executeSelect(" +
+                "'select ''Foo'' as t_varchar, 1::integer as t_integer, " +
+                "1.5::float as t_float, 23.67::decimal(8,2) as t_decimal, " +
+                "''2005-06-01''::date as t_date, ''20:56''::time as t_time, " +
+                "''2006-02-04 23:55:10''::timestamp as t_timestamp')";
+
+			rs = stmt.executeQuery(sql);
+			System.out.println("SQL = " + sql);
+			System.out.println("results:");
+			while(rs.next())
+			{
+				System.out.println(rs.getString(1));
+			}
+			rs.close();
+		}
+		finally
+		{
+			if(rs != null)
+			{
+				try { rs.close(); } catch(SQLException e) {}
+				rs = null;
+			}
+		}
+	}
 }
